/*
 * MP3 over RTP streaming server. Open a TCP socket on port 8000 or 8001, get
 * a Shoutcast icycast or audiocast connection with a valid password, and
 * then receive an MP3 stream on this socket. Then read in MP3 frames, and
 * convert them into a sequence of RTP packets which are then multicast.
 * 
 * Some code borrowed from Obsequiem http://obs.freeamp.org, but converted to C
 * and slightly rewritten to act as a server. The rewrite was done by Warren
 * Toomey, 2002.
 * 
 * I wouldn't say that the code was the prettiest or cleanest. It certainly
 * needs to be more consistent w.r.t names of identifiers. It appears to be
 * reasonably solid. Feel free to clean it up and improve functionality!
 * 
 * This code is covered by the GNU GENERAL PUBLIC LICENSE Version 2.
 * 
 * $Revision: 1.32 $
 * 
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <unistd.h>
#include <fcntl.h>
#include <syslog.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <time.h>
#include <sys/time.h>
#include <sys/un.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <errno.h>

#define BSIZE 2048		/* Fixed-sized bufs for reading strings */
#define PROTO_AUDIOCAST 1
#define PROTO_ICYCAST	2

typedef struct RTPHeader {
  int flags;
  int timestamp;
  int syncsource_id;
  int audioheader;
}   RTPHeader;

/* Configuration information */
char *ConfigFile= "/etc/shout2rtprc";
char *Password= NULL;
char *URL= NULL;
char *StreamTitle= NULL;
char *ShoutAddr= NULL;
char *RTPAddr= NULL;
char *DeadAirDir= NULL;
char *MulticastAddr= "225.2.1.5";
int MulticastPort= 4420;
int Ttl=1;
int ShoutProtocol= 0;		/* ==1 is audiocast, ==2 is icycast */
int BecomeDaemon= 0;
int DeadAirDelay= 60;

/* Addresses and other stuff */
const int PacketSize= 1500;
struct sockaddr_in sSockAddr, sTitleAddr, sShoutAddr;
struct sockaddr_in Fred;
socklen_t addrlen= sizeof(Fred);
int SSocket= 0, Socket= 0, TitleSocket= 0, ShoutSocket= 0;
int iBitRate, iBaseTime;
time_t lasttime, thistime;
pid_t deadair_pid= 0;
int we_are_deadair_child= 0;


/* Code to deal with the list of dead air files */
#define MAX_DEADAIRFILES	20
char *DeadAir[MAX_DEADAIRFILES];
int numDeadAirFiles= 0;

/*
 * Find and build the list of dead air files. Yes, I know I should use
 * opendir/readdir, but this is easier :)
 */
void find_deadair_files(void)
{
  FILE *zin;
  char *cptr, buf[BSIZE];

  if (DeadAirDir == NULL) return;

  snprintf(buf, BSIZE - 2,
       "/bin/ls -1 %s/*.mp3 %s/*.MP3 2> /dev/null", DeadAirDir, DeadAirDir);

  zin= popen(buf, "r");
  if (zin == NULL) { syslog(LOG_ERR, "deadairdir pipe open: %m\n"); return; }
  while (!feof(zin) && (numDeadAirFiles < MAX_DEADAIRFILES)) {
    cptr= fgets(buf, BSIZE - 2, zin);
    if (cptr == NULL) break;
    /* Trim the newline from the end of the string */
    buf[strlen(buf) - 1]= '\0';

    DeadAir[numDeadAirFiles]= strdup(buf);
    numDeadAirFiles++;
  }
  pclose(zin);
  srandom(time(NULL));
}


char * choose_random_deadair_file(void)
{
  int rndnum;

  if (numDeadAirFiles == 0) return (NULL);
  if (numDeadAirFiles == 1) return (DeadAir[0]);
  rndnum= random() % numDeadAirFiles;
  return (DeadAir[rndnum]);
}

/* Internal read function, which
 * doesn't give up on EAGAIN errors
 */
int Read(int d, void *buf, size_t nbytes)
{
 int result;

 while (1) {
   result= read(d, buf, nbytes);
   /* Exit loop if result >=0 or errno is not EAGAIN */
   if ((result>=0) || (errno!=EAGAIN)) break;
   syslog(LOG_ERR, "EAGAIN error in stream read");
 }
 return(result);
}


/*
 * Open and read the configuration file so that we know the stream's title,
 * the URL and our password. Return 1 if ok, or 0 if failure.
 */
int ReadConfigFile(void)
{
  FILE *zin;
  char buf[BSIZE];
  char *cptr;
  char *varname, *value;

  zin= fopen(ConfigFile, "r");
  if (zin == NULL) return (0);

  while (!feof(zin)) {
    cptr= fgets(buf, BSIZE - 2, zin);
    if (cptr == NULL) break;

    if (buf[0] == '#') continue;

    /* Trim the newline from the end of the string */
    buf[strlen(buf) - 1]= '\0';

    /*
     * Find the= sign in the line, and NULL it after setting value to point
     * after it.
     */
    value= strchr(buf, '=');
    if (value == NULL) {
      fprintf(stderr, "No= sign on this line in config file:\n\t%s\n", buf);
      return (0);
    }
    *value= '\0'; value++;

    /* Walk value up past any leading spaces */
    while ((*value == ' ') || (*value == '\t')) value++;

    /* Trim any trailing whitespace off the variable name */
    varname= buf;
    cptr= strpbrk(varname, " \t");
    if (cptr != NULL) *cptr= '\0';

    if (!strcmp(varname, "title"))	{ StreamTitle= strdup(value);continue; }
    if (!strcmp(varname, "URL"))	{ URL= strdup(value); continue; }
    if (!strcmp(varname, "password"))	{ Password= strdup(value); continue; }
    if (!strcmp(varname, "rtpaddr"))	{ RTPAddr= strdup(value); continue; }
    if (!strcmp(varname, "shoutaddr"))	{ ShoutAddr= strdup(value); continue; }
    if (!strcmp(varname, "deadairdir")) { DeadAirDir= strdup(value); continue; }
    if (!strcmp(varname,"multicastaddr")){MulticastAddr=strdup(value);continue;}
    if (!strcmp(varname, "deadairdelay")){ DeadAirDelay= atoi(value);continue; }
    if (!strcmp(varname, "daemon")) { BecomeDaemon= atoi(value); continue; }
    if (!strcmp(varname, "multicastport")){MulticastPort= atoi(value);continue;}
    if (!strcmp(varname, "ttl"))       { Ttl= atoi(value);continue; }
    if (!strcmp(varname, "shoutprotocol")) {
      if (!strcmp(value, "audiocast")) ShoutProtocol= PROTO_AUDIOCAST;
      if (!strcmp(value, "icycast")) ShoutProtocol= PROTO_ICYCAST;
    }
  }

  /* Error checking */
  if (Password == NULL) {
    fprintf(stderr, "No password in config file\n"); exit(1);
  }
  fclose(zin); return (1);
}



/*
 * Construct the TCP socket which we use to get a connection from the
 * Shoutcast stream generator. Returns 1 if nothing went wrong, or 0 if we
 * failed. The Port is the port to use. If Addr is non-null, this is the IP
 * address to bind to. If NULL, we bind to all our IP addresses.
 */
int MakeStreamSocket(char *BindAddr, int Port)
{
  /* Create socket */
  SSocket= socket(AF_INET, SOCK_STREAM, 0);
  if (SSocket < 0) { syslog(LOG_ERR, "stream socket open: %m\n"); return (0); }

  /* Name socket using wildcards */
  sShoutAddr.sin_family= AF_INET;
  sShoutAddr.sin_port= htons(Port);

  if (BindAddr != NULL) sShoutAddr.sin_addr.s_addr= inet_addr(BindAddr);
  else sShoutAddr.sin_addr.s_addr= INADDR_ANY;

  if (bind(SSocket, (struct sockaddr *)&sShoutAddr, sizeof(sShoutAddr))) {
    syslog(LOG_ERR, "stream socket bind: %m\n"); return (0);
  }
  return (1);
}

/*
 * Get the stream in Shoutcast format from the mixer. Returns 1 if we got
 * connected, or 0 if some catastrophic error. The code loops until there is
 * a validated stream connection.
 */
int GetShoutStream(void)
{
  char ch, buf[BSIZE], *src= "SOURCE ", *dead;
  int err, i, nlcount= 0, flags;
  struct timeval sTime;

  /* Get connection from stream transmitter */
  listen(SSocket, 1); lasttime= time(NULL);

  /* Don't block if there are no connections */
  flags= fcntl(SSocket, F_GETFL, NULL);
  if (flags != -1) { flags |= O_NONBLOCK; fcntl(SSocket, F_SETFL, flags); }

  while (1) {
    ShoutSocket= accept(SSocket, (struct sockaddr *)&Fred, &addrlen);

    if (ShoutSocket == -1) {
      /*
       * No Shoutcast stream yet. If any dead air files exist and it is >=
       * DeadAirDelay, then play a random dead air file. Otherwise, sleep for
       * 2 seconds and wait for another connection.
       */
      thistime= time(NULL); dead= choose_random_deadair_file();

      if (dead && ((thistime - lasttime) >= DeadAirDelay)) {
	if (deadair_pid > 0) {
	  waitpid(deadair_pid, &i, 0);
	  syslog(LOG_INFO,"Parent reaped pid %d to redo deadair\n",deadair_pid);
	}
	/*
	 * Firstly spawn a child to play the dead air file, so that the
	 * parent can go back and try to accept again.
	 */
	deadair_pid= fork();
	if (deadair_pid == 0) {
	  ShoutSocket= open(dead, O_RDONLY);
	  if (ShoutSocket != -1) {
	    syslog(LOG_INFO, "Opening dead air file %s\n", dead);
	    we_are_deadair_child= 1; return (1);
	  } else exit(1);
	} else lasttime= thistime;
      }
      sleep(2); continue;
    }
    syslog(LOG_INFO, "Connection from Shoutcast client at %s\n",
	   inet_ntoa(Fred.sin_addr));


    /*
     * We have a TCP connection from a potential Shoutcast stream. Turn off
     * non-blocking now.
     */
    flags &= (~O_NONBLOCK);
    fcntl(SSocket, F_SETFL, flags);
    fcntl(ShoutSocket, F_SETFL, flags);

    /*
     * Set a maximum of 15 seconds inactivity on reads from the socket. If
     * the transmitter goes away, this allows us to quickly recover.
     */
    sTime.tv_sec= 15; sTime.tv_usec= 0;
    err= setsockopt(ShoutSocket, SOL_SOCKET, SO_RCVTIMEO, (void *)&sTime,
		     sizeof(sTime));
    if (err != 0) syslog(LOG_ERR, "setting read inactivity to 15 secs: %m\n");
    /*
     * Now start reading the connection information from the transmitter. We
     * are only really interested in the password. Once we get a blank line
     * from the sender, we can send Ok and we are done. This is terribly
     * inefficient, but it works.
     */
    if (ShoutProtocol == PROTO_AUDIOCAST) {
      err= Read(ShoutSocket, &buf, 7);
      if (err != 7) { close(ShoutSocket); continue; }
      err= memcmp(buf, src, 7);
      if (err) { close(ShoutSocket); continue; }
    }
    /* Now the password */
    for (i= 0; i < BSIZE - 2; i++) {
      err= Read(ShoutSocket, &buf[i], 1);
      if (err != 1) {
	syslog(LOG_ERR, "password read error: %m\n");
	close(ShoutSocket); continue;
      }
      if ((buf[i] == '\n') || (buf[i] == ' ')) break;
    }
    buf[i]= '\0';
    if (strncmp(buf, Password, strlen(Password))) {
      syslog(LOG_ERR, "bad password from Shoutcast client\n");
      close(ShoutSocket); continue;
    }
    if (ShoutProtocol == PROTO_AUDIOCAST) write(ShoutSocket, "OK\n", 3);
    else write(ShoutSocket, "OK2\r\nicy-caps:11\r\n\r\n", 20);

    /* Now read up to 2 \ns */
    while (1) {
      err= Read(ShoutSocket, &ch, 1);
      if (err != 1) { close(ShoutSocket); continue; }
      if (ch == '\n') nlcount++;
      else nlcount= 0;
      if (nlcount == 2) break;
    }
    syslog(LOG_INFO, "Connection from Shoutcast stream established\n");

    /*
     * If there is a dead air child, wait for it to terminate before we start
     * reading the stream from the Shoutcast transmitter.
     */
    if (deadair_pid > 0) {
      syslog(LOG_INFO, "Parent trying to reap pid %d\n", deadair_pid);
      waitpid(deadair_pid, &i, 0);
      syslog(LOG_INFO, "Parent reaped pid %d\n", deadair_pid);
    }
    deadair_pid= 0; return (1);
  }
}


int InitMultiSocket(int *sock, struct sockaddr_in *SockAddr, const char *szAddr,
		      const char *szLocalAddr, int Port, int ttl)
{
  int ret, iLoop= 1;
  struct sockaddr_in stLocal;
  char cttl= (char)ttl;
  char cLoop= 0;

  if  ((*sock) > 0) return (1);

  *sock= socket(AF_INET, SOCK_DGRAM, 0);
  if  ((*sock) < 0) { syslog(LOG_ERR, "UDP socket() failed: %m\n"); return 0; }

  cLoop= 1;
  SockAddr->sin_family= AF_INET;
  SockAddr->sin_port= htons(Port);
  SockAddr->sin_addr.s_addr= inet_addr(szAddr);

  ret= setsockopt((*sock), SOL_SOCKET, SO_REUSEADDR, &iLoop, sizeof(int));
  if (ret < 0) {
      syslog(LOG_ERR, "setsockopt SO_REUSEADDR failed: %m\n");
      return 0;
  }
  ret= setsockopt((*sock), IPPROTO_IP, IP_MULTICAST_TTL, &cttl, sizeof(char));
  if (ret < 0) {
    syslog(LOG_ERR, "setsockopt IP_MULTICAST_TTL failed: %m\n"); return 0;
  }
  ret= setsockopt((*sock), IPPROTO_IP, IP_MULTICAST_LOOP, &cLoop, sizeof(char));
  if (ret < 0) {
    syslog(LOG_ERR, "setsockopt IP_MULTICAST_LOOP failed: %m\n"); return 0;
  }

  /* If we have a local addr, bind to that specific address */
  if (szLocalAddr && strlen(szLocalAddr) > 0) {
    memset(&stLocal, 0, sizeof(stLocal));
    stLocal.sin_addr.s_addr= inet_addr(szLocalAddr);
    ret= setsockopt((*sock), IPPROTO_IP, IP_MULTICAST_IF,
		 &stLocal.sin_addr.s_addr, sizeof(stLocal.sin_addr.s_addr));
    if (ret < 0) {
      syslog(LOG_ERR, "binding UDP to specific IP failed: %m\n"); return 0;
    } else syslog(LOG_INFO, "binding UDP to IP address %s ok\n", szLocalAddr);
  }
  return(1);
}


long long GetTime_ll(void)
{
  struct timeval sTime;

  gettimeofday(&sTime, NULL);
  if (iBaseTime == 0) iBaseTime= sTime.tv_sec;

  return (((long long)sTime.tv_sec - (long long)iBaseTime) * 100000) +
    ((long long)sTime.tv_usec / 10);
}

void ResetTime(void)
{
  struct timeval sTime;

  gettimeofday(&sTime, NULL); iBaseTime= sTime.tv_sec;
}

int bitrate_table[2][4][16] = {
  /* MPEG II */
  {
    /* reserved */
    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
    /* layer 3 */
    {0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160, 0},
    /* layer 2 */
    {0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160, 0},
    /* layer 1 */
    {0, 32, 48, 56, 64, 80, 96, 112, 128, 144, 160, 176, 192, 224, 356, 0}
  },
  /* MPEG I */
  {
    /* reserved */
    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
    /* layer 3 */
    {0, 32, 40, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320, 0},
    /* layer 2 */
    {0, 32, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320, 384, 0},
    /* layer 1 */
    {0, 32, 64, 96, 128, 160, 192, 224, 256, 288, 320, 352, 384, 416, 448, 0}
  }
};

int samplerate_table[2][3] = {
  {22050, 24000, 16000},
  {44100, 48000, 32000}
};


/*
 * Read and decode the first 3 octets from an MP3 frame to work out frame
 * size, bitrate and samplerate. If dontread==1, use the previous values.
 * Returns 1 if ok, 0 if an error. Copy the 3 bytes into pBuffer if ok.
 */

static unsigned char fheader[3];

int GetMPEGFrameInfo_b(unsigned char *pBuffer, int dontread, int *iFrameSize,
		   int *iBitrate, int *iSamplerate)
{
  int err;
  int mpeg;
  int layer;
  int bitrate_index;
  int samplerate_index;
  int pad;
  int no_valid_header=1;

  while (no_valid_header) {
    if (!dontread) {
      fheader[0]= 0;
      while (1) {
        err= Read(ShoutSocket, &fheader[1], 1);
        if (err != 1) { syslog(LOG_INFO, "Read failure #1\n"); return (0); }
        if ((unsigned char)fheader[0] == 0xff &&
	    ((unsigned char)fheader[1] & 0xf0) == 0xf0)
	  break;
        fheader[0]= fheader[1];
      }
      err= Read(ShoutSocket, &fheader[2], 1);
      if (err != 1) { syslog(LOG_INFO, "Read failure #2\n"); return (0); }
    }
    mpeg= (fheader[1] & 0x08) >> 3;
    layer= (fheader[1] & 0x06) >> 1;
    bitrate_index= (fheader[2] & 0xf0) >> 4;
    samplerate_index= (fheader[2] & 0x0c) >> 2;
    pad= (fheader[2] & 0x02) >> 1;

    *iBitrate= bitrate_table[mpeg][layer][bitrate_index];
    *iSamplerate= samplerate_table[mpeg][samplerate_index];
  
    /* Exit the outside loop if we now have a valid header */
    if ((*iBitrate != 0) && (*iSamplerate != 0)) break;
  
    syslog(LOG_ERR, "Frame err: mpeg %d layer %d bit %d sample %d pad %d\n",
	   mpeg, layer, bitrate_index, samplerate_index, pad);
    dontread=0; 
  }

  *iFrameSize= (144000 * *iBitrate / *iSamplerate) + pad;

  pBuffer[0]= fheader[0]; pBuffer[1]= fheader[1]; pBuffer[2]= fheader[2];
  return (1);
}

int leftover_frame= 0;
/*
 * Construct a packet by reading in as many MP3 frames as we can until there
 * is no more room in the buf. Return the number of frames read in. If we
 * return 0, then the connection to the Shoutcast server has gone away.
 */
int MakePacket(unsigned char *buf, const int pktsize, int *iBytesInPacket,
	   int *iSampleRate)
{
  int iFrameSize, iNumFrames= 0, iBitRate;
  int err;
  int numread;

  *iBytesInPacket= 0;
  while (1) {
    err= GetMPEGFrameInfo_b(buf, leftover_frame, &iFrameSize, &iBitRate,
			     iSampleRate);
    if (err == 0) {
      if (BecomeDaemon==0) fprintf(stderr, "1st Shoutcast read failed\n");
      syslog(LOG_ERR, "1st Shoutcast read failed\n");
      leftover_frame= 0; return (0);
    }
    if ((*iBytesInPacket + iFrameSize) > pktsize) {
      if ((BecomeDaemon==0) && (iNumFrames==0))
        fprintf(stderr, "Can't fit %d and %d into %d\n",
				*iBytesInPacket, iFrameSize, pktsize);
      leftover_frame= 1; return (iNumFrames);
    }

    /* We have enough room to read in the rest of the frame */
    leftover_frame= 0;
    (*iBytesInPacket) += iFrameSize;
    buf += 3; iFrameSize -= 3;
    while (iFrameSize > 0) {
      numread= Read(ShoutSocket, buf, iFrameSize);
      if (numread <= 0) {
	/* A real read failure, give up on the read */
	syslog(LOG_ERR, "2nd Shoutcast read failed: %s\n",strerror(errno));
	if (BecomeDaemon==0)
	  fprintf(stderr, "2nd Shoutcast read failed: %s\n", strerror(errno));
	leftover_frame= 0; return (0);
      }
      iFrameSize -= numread; buf += numread;
    }
    iNumFrames++;
  }
   if (BecomeDaemon==0)
      fprintf(stderr, "How did we get here with %d frames?\n",iNumFrames);
  return (iNumFrames);
}

/*
 * If this ever returns, it is because we can't read from the Shoutcast
 * transmitter
 */
void WorkerThread(void)
{
  unsigned char pBuffer[PacketSize];
  RTPHeader *pPacket;
  unsigned int uPacketNum, time_between_packets;
  int packet_delay;
  long long llPacketTime= 0, llCurrentTime;
  unsigned uSequenceNumber;
  int ret, iSampleRate;
  int iBytesCopied;
  int iFlagMask;
  int frames_in_pkt= 0;
  int hdrlen= sizeof(RTPHeader);

  pPacket= (RTPHeader *) pBuffer;
  iFlagMask= 0x800E << 16;
  srand(getpid());
  pPacket->syncsource_id= htonl(rand());
  pPacket->audioheader= 0;
  packet_delay= 0;

  for (uPacketNum=0, uSequenceNumber=rand();; uPacketNum++, uSequenceNumber++) {

    /* Make a packet. If we can 't make a packet for some reason, give up */
    frames_in_pkt= MakePacket(&pBuffer[hdrlen],
	  sizeof(pBuffer) - sizeof(RTPHeader), &iBytesCopied, &iSampleRate);

    if (frames_in_pkt <= 0) {
      fprintf(stderr, "Can't make a packet, giving up\n"); 
      syslog(LOG_ERR, "Can't make a packet, giving up\n"); break;
    }

    pPacket->flags= htonl(iFlagMask | (uSequenceNumber & 0xFFFF));

    /* iTimeBetweenPackets is in 10us units */
    time_between_packets= ((1000 * 1152) / (iSampleRate / 100)) * frames_in_pkt;

    /* uPacketTime is in 10us units */
    llPacketTime += (long long)time_between_packets;
    pPacket->timestamp= 0;

    if (uPacketNum == 0) ResetTime();

    llCurrentTime= GetTime_ll();
    packet_delay= (int)(llPacketTime - llCurrentTime);
    if (packet_delay < 0) packet_delay= 0;

    usleep(packet_delay * 10);

    ret= sendto(Socket, pBuffer, iBytesCopied + sizeof(RTPHeader), 0,
		 (struct sockaddr *)&sSockAddr, sizeof(struct sockaddr_in));
    if (ret < 0) {
      syslog(LOG_ERR, "sendto() returned: %m\n"); break;
    }
  }
  syslog(LOG_INFO, "RTPTransmitter thread failed.\n");
  return;
}

void TitleThread()
{
  char oTitle[2048];
  int ret;

  snprintf(oTitle, 2040,
	   "x-audiocast-streamtitle: %s\nx-audiocast-streamurl: %s\n",
	   StreamTitle, URL);

  for (;;) {
    ret= sendto(TitleSocket, oTitle, strlen(oTitle), 0,
		 (struct sockaddr *)&sTitleAddr, sizeof(sTitleAddr));

    if (ret < 0) { syslog(LOG_ERR, "title sendto() returned: %m\n"); exit(1); }
    sleep(10);
  }
  syslog(LOG_INFO, "Title thread exiting\n");
  exit(0);
}

int main(int argc, char *argv[])
{
  pid_t pid;
  int err;

  /* Read config file */
  err= ReadConfigFile();
  if (err != 1) {
    printf("Bad or missing config file at %s, giving up\n", ConfigFile);exit(1);
  }
  find_deadair_files();

  if (ShoutProtocol == 0) {
    fprintf(stderr, "No Shoutcast protocol selected, using audiocast\n");
    ShoutProtocol= PROTO_AUDIOCAST;
  }
  if (StreamTitle == NULL) {
    StreamTitle= "Unknown RTP transmission\n";
    fprintf(stderr, "No title in config file, using %s\n", StreamTitle);
  }
  if (URL == NULL) {
    URL= "http://minnie.tuhs.org\n";
    fprintf(stderr, "No URL in config file, using %s\n", URL);
  }

  /* Become a daemon, and open up the syslog */
  if (BecomeDaemon) daemon(0, 0);

  openlog("shout2rtp", 0, LOG_LOCAL5);
  syslog(LOG_INFO, "Starting operations, $Revision: 1.32 $\n");

  /* Make the socket for the Shoutcast stream */
  if (ShoutProtocol == PROTO_AUDIOCAST) err= MakeStreamSocket(ShoutAddr, 8000);
  else err= MakeStreamSocket(ShoutAddr, 8001);
  if (err != 1) {
    syslog(LOG_INFO, "Couldn't create Shoutcast socket, exiting\n"); exit(1);
  }

  /* Make the title output socket */
  err= InitMultiSocket(&TitleSocket, &sTitleAddr,
			MulticastAddr, RTPAddr, MulticastPort + 2, Ttl);
  if (err != 1) {
    syslog(LOG_INFO, "Couldn't create title socket, exiting\n"); exit(1);
  }

  /*
   * At this point, we split into two processes. The child will act as the
   * process that multicasts the title of the RTP stream, and the server will
   * multicast the stream itself.
   */
  pid= fork();
  switch (pid) {
  case 0:			/* Child process, title transmitter */
    /* Close unwanted sockets */
    close(SSocket); close(Socket); close(ShoutSocket);
    TitleThread();		/* This should never return */
    exit(1);
  }

  /* Close the unwanted title socket */
  close(TitleSocket);

  /* We can now create our RTP output socket */
  err= InitMultiSocket(&Socket, &sSockAddr,
			MulticastAddr, RTPAddr, MulticastPort, Ttl);
  if (err != 1) {
    syslog(LOG_INFO, "Couldn't create RTP socket, exiting\n"); exit(1);
  }

  /* Loop forever getting (re)connected to the Shoutcast stream */
  /* and doing the Shoutcast to RTP conversion */
  while (1) {
    err= GetShoutStream();
    if (err != 1) {
      syslog(LOG_INFO, "Error in Shoutcast connection, exiting\n"); exit(1);
    }

    /* Start the converter. This only returns */
    /* when either the RTP socket stops working, */
    /* or when we lose connection from the Shoutcast */
    /* stream generator */
    WorkerThread();
    syslog(LOG_INFO, "Lost connection from Shoutcast stream\n");
    close(ShoutSocket);

    /* If we are a deadair playing child, then exit here */
    if (we_are_deadair_child) {
      syslog(LOG_INFO, "Dead air child dying now\n"); exit(0);
    }
  }
  exit(0);
}
